// LOGGER.C
// Logging and function entry/return tracking functions
#define _LOGGER_C_
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include <stdtypes.h>
#include <logger.h>
#include <extant.h>
#include <misc.h>

// Thinking that this should never be too deep, but handling fields 
// dynamically might become major source of memory fragmentation or 
// performance suckage.  Therefore, using static strings.
typedef struct flog_tag
  {
    word32 flags;
    char   filename[128];
    char   funcname[64];
    char   dbstr[64];
    ptrint dbdata;
    int    dbdtype;
    int    lineno;
    ptrint state[4];
    // FIXME -- add timestamping information
  } flog_t;

static FILE   *logfile = NULL;
static xst    *flog    = NULL;
static flog_t *frec    = NULL;  // points at 'top' flog record

void function_entry ( char *filename, int    lineno, char *funcname, 
                      char *s,        ptrint d,      char *type )
  {
    flog_t  fr;
    if ( flog == NULL )
      {
        int res;
        flog = (xst*) malloc ( sizeof(xst) );
        if ( flog == NULL ) return;
        res = xst_init ( flog, sizeof(flog_t), 4 ); // FIXME is this right?
        if ( res != 0 ) return;
        // FIXME -- also catch SEGV
      }
    fr.flags = 1;
    if ( filename != NULL )
      {
         strncpy ( fr.filename, filename, 127 );
         fr.filename[127] = 0; // strncpy is not guaranteed to null-terminate
      } 
    else fr.filename[0] = 0;
    if ( funcname != NULL )
      {
         strncpy ( fr.funcname, funcname, 63 );
         fr.funcname[63] = 0; // strncpy is not guaranteed to null-terminate
      } 
    else fr.funcname[0] = 0;
    fr.lineno = lineno;
    if ( s != NULL )
       {
         strncpy ( fr.dbstr, s, 63 );
         fr.dbstr[63] = 0; // strncpy is not guaranteed to null-terminate
       } 
    else fr.dbstr[0] = 0;
    if ( type == NULL ) fr.dbdtype = 0;
    else 
      {
        if      ( 0 == strcmp ( "i", type ) ) fr.dbdtype = 1; // int 
        else if ( 0 == strcmp ( "s", type ) ) fr.dbdtype = 2; // str
        else fr.dbdtype = -1;
      }
    switch ( fr.dbdtype )
      {
        case  1 : fr.dbdata = (int)d; break;
        case  2 : fr.dbdata = (ptrint)d; break;
        default : fr.dbdata = d; break;
      }
    xst_add ( flog, &fr );
    frec = (flog_t*)(xst_fetchp ( flog, (flog->n)-1 ));
  } // END of function function_entry()


void function_extra ( char *filename, int    lineno, char *funcname, 
                      char *s,        ptrint d,      char *type )
  {
    flog_t  fr;
    if ( flog == NULL )
      {
        int res;
        flog = (xst*) malloc ( sizeof(xst) );
        if ( flog == NULL ) return;
        res = xst_init ( flog, sizeof(flog_t), 4 ); // FIXME is this right?
        if ( res != 0 ) return;
      }
    fr.flags = 3;
    if ( filename != NULL )
      {
         strncpy ( fr.filename, filename, 127 );
         fr.filename[127] = 0; // strncpy is not guaranteed to null-terminate
      } 
    else fr.filename[0] = 0;
    if ( funcname != NULL )
      {
         strncpy ( fr.funcname, funcname, 63 );
         fr.funcname[63] = 0; // strncpy is not guaranteed to null-terminate
      } 
    else fr.funcname[0] = 0;
    fr.lineno = lineno;
    if ( s != NULL )
       {
         strncpy ( fr.dbstr, s, 63 );
         fr.dbstr[63] = 0; // strncpy is not guaranteed to null-terminate
       } 
    else fr.dbstr[0] = 0;
    if ( type == NULL ) fr.dbdtype = 0;
    else 
      {
        if      ( 0 == strcmp ( "i", type ) ) fr.dbdtype = 1; // int 
        else if ( 0 == strcmp ( "s", type ) ) fr.dbdtype = 2; // str
        else fr.dbdtype = -1;
      }
    switch ( fr.dbdtype )
      {
        case  1 : fr.dbdata = ((int*)d)[0]; break;
        case  2 : fr.dbdata = (ptrint)d; break;
        default : fr.dbdata = d; break;
      }
    xst_add ( flog, &fr );
    frec = (flog_t*)(xst_fetchp ( flog, (flog->n)-1 ));
  } // END of function function_extra()

void function_exit()
  {
    flog_t *fr = (flog_t*)(flog->d);
    if ( flog->n < 0 ) { flog->n = 0; return; }
    // first, get rid of "extra" records
    while ( ( flog->n > 0              ) && 
            ( ( fr[flog->n].flags | 2 ) == 1 )    )
      flog->n--;
    // now, if safe, get rid of "entry" record.
    if ( flog->n > 0 ) flog->n--;
  }

void log_exit(void)
  {
    if ( logfile != NULL ) 
      { 
        LOG ( "process exiting" );
        fclose ( logfile );
      } else {
        fprintf ( stderr, "process exiting\n" );
      }
    return;
  }

void log_init()
  {
    char fname[32];
    sprintf ( fname, "/tmp/log.%d", getpid() );
    logfile = fopen ( fname, "w" );
    if ( logfile == NULL ) { fprintf ( stderr, "cannot open logfile '%s'; aborting\n", fname ); exit(1); }
    atexit ( log_exit );
    return;
  }

void llog ( char *filename, char *funcname, int linenum, char *s )
  {
    char curdate[16];
    time_t curt;
    if ( s == NULL ) return;
    curt = time(NULL);
    if ( logfile == NULL ) log_init();
    curdate[0] = 0;
    strftime ( curdate, 63, "%Y %m %d %a", localtime(&curt) );
    fprintf ( logfile, "[%s] %u (%s:%s:%d) %s\n", curdate, curt, filename, funcname, linenum, s );
    fflush ( logfile );
    return;
  }

void errortrap ( char *s, int exitprocess )
  {
    char    buf[1024];
    char   *ts = NULL;
    int     i, k;
    flog_t *fr;

    FENTRY ( s, exitprocess, "i" );
    if ( s == NULL )
      { 
        LOG ( mkstr ( buf, "ERROR! (%d)", exitprocess ) );
      } else {
        LOG ( s );
      }
    LOG ( "Function stack dump follows:" );
    for ( i = (flog->n)-1; i >= 0; i-- )
      {
        fr = xst_fetchp(flog,i);
        if ( fr == NULL ) continue; // zzapp -- right syntax? damn perl!
        ts = fr->dbstr;
        switch ( fr->dbdtype )
          { // FIXME -- replace consts with macros with meaningful names
            case 0 : LOG ( mkstr( buf, "%3d: (%s:%s:%d) %s (null) states: 0x%X 0x%X 0x%X 0x%X", i, fr->filename, fr->funcname, fr->lineno, ts, fr->state[0], fr->state[1], fr->state[2], fr->state[3] ) ); break;
            case 1 : LOG ( mkstr( buf, "%3d: (%s:%s:%d) %s '%d' states: 0x%X 0x%X 0x%X 0x%X",   i, fr->filename, fr->funcname, fr->lineno, ts, (int)(fr->dbdata), fr->state[0], fr->state[1], fr->state[2], fr->state[3] ) ); break;
            case 2 : LOG ( mkstr( buf, "%3d: (%s:%s:%d) %s '%s' states: 0x%X 0x%X 0x%X 0x%X",   i, fr->filename, fr->funcname, fr->lineno, ts, (char*)(fr->dbdata), fr->state[0], fr->state[1], fr->state[2], fr->state[3] ) ); break;
            case 3 : LOG ( mkstr( buf, "%3d: (%s:%s:%d) %s '0x%X' states: 0x%X 0x%X 0x%X 0x%X", i, fr->filename, fr->funcname, fr->lineno, ts, (unsigned int)(fr->dbdata), fr->state[0], fr->state[1], fr->state[2], fr->state[3] ) ); break;
            default: LOG ( mkstr( buf, "%3d: (%s:%s:%d) %s (unknown type %d) '0x%X' states: 0x%X 0x%X 0x%X 0x%X", i, fr->filename, fr->funcname, fr->lineno, ts, fr->dbdtype, (unsigned int)(fr->dbdata), fr->state[0], fr->state[1], fr->state[2], fr->state[3] ) ); break;
	  } // end of switch
      } // end of for
    if ( exitprocess == 1 ) exit(1);
    FEXIT
  } // END of function errortrap()

