// Regular Expression library wrapper for PCRE -- easier to use, with 
// compiled expression caching and function profiling.
#define _RE_C_
#include <string.h>
#include <pcre.h>
#include "stdtypes.h"
#include "extant.h"
#include "logger.h" 
#include "misc.h"
#include "re.h"
#warning "Ignore the cast errors -- stdtypes.h is designed to guarantee that the ptrint type is safe"

#ifndef  REGEX_CACHE_SLOTS
#define  REGEX_CACHE_SLOTS  (1023)
#endif

#ifndef  REGEX_DEFAULT_REPLACEMENT_POLICY
#define  REGEX_DEFAULT_REPLACEMENT_POLICY  (0)
#endif

#ifndef  REGEX_DEFAULT_POSTCOMPILE
#define  REGEX_DEFAULT_POSTCOMPILE (1)
#endif

#ifndef  REGEX_N_VECTORS
#define  REGEX_N_VECTORS (128)
#endif

typedef struct rexc_type
  {
    char       *pattern;
    word32      hashvalue;
    pcre       *func;
    pcre_extra *postcompile;
  } rexc_t;

static xst *cache = NULL;  // extant of rexc_t
static int  cache_replacement_policy = 0;
static int  regex_postcompile = 1;
static int  debugging = 0;

static void re_cache_init()
  {
    int i, k, res;
    FENTRY ( "initializing cache", NULL, NULL );
    if ( debugging > 1 ) fprintf ( stderr, "initializing cache\n" );
    cache = malloc ( sizeof(xst) );
    if ( cache == NULL ) { ERROR ("cannot allocate memory for cache",0); FEXIT return; }
    res = xst_init ( cache, sizeof(rexc_t), REGEX_CACHE_SLOTS ); // FIXME -- check for errors
    cache->n = cache->max;
    for ( i = 0;  i < REGEX_CACHE_SLOTS; i++ )
      {
        rexc_t *p = (rexc_t*)(cache->d);
        p[i].pattern = NULL;
        p[i].hashvalue = 0;
        p[i].func = NULL;
        p[i].postcompile = NULL;
      } 
    if ( debugging > 1 ) fprintf ( stderr, "initialization done\n" );
    FEXIT return;
  } // end of function re_cache_init()

int re ( char *pattern, char *s, xpt *x )
  { 
    int      gotmatch = 0;
    rexc_t  *rec;
    word32   hval;
    const char *err;
    int      erroffset;
    int      vectors[REGEX_N_VECTORS*3];

    FENTRY ( "applying pattern", pattern, "s" );
    FEXTRA ( "data string", s, "s" ); // zzapp
    if ( debugging > 1 ) fprintf ( stderr, "re() entered\n" );
    if ( pattern == NULL ) { FEXIT  return ( -1 ); }
    if ( s       == NULL ) { FEXIT  return ( -2 ); }
    if ( strlen ( pattern ) > 32767 ) { FEXIT  return ( -3 ); }
    if ( strlen ( s       ) > 32767 ) { FEXIT  return ( -4 ); }
    // first, initialize the cache if needed
    if ( cache == NULL ) re_cache_init();
    if ( cache == NULL ) { FEXIT  return ( -5 ); }
    // now, find our cache line and, if needed, put our stuff in it
    hval = hasher ( pattern );
    if ( debugging > 1 ) fprintf ( stderr, "hash value of '%s' is %u\n", pattern, hval );
    rec = (rexc_t*) xst_fetchp ( cache, hval % REGEX_CACHE_SLOTS );
    if ( rec == NULL ) { ERROR ("bad cache reference",1); FEXIT return ( -9 ); }
    if ( debugging > 1 ) fprintf ( stderr, "slot %u maps to cache line { 0x%X %u 0x%X 0x%X } (rp %d)\n", ( hval % REGEX_CACHE_SLOTS ), rec->pattern, rec->hashvalue, rec->func, rec->postcompile, cache_replacement_policy );
    if ( ( rec->pattern   == NULL ) || // empty cache line
         ( rec->hashvalue != hval ) || // mismatching hash values
         ( ( rec->hashvalue == hval ) &&
           ( cache_replacement_policy == 0 ) &&
           ( strcmp ( rec->pattern, pattern ) != 0 ) ) ) // pattern mismatch
      {
        // FROM THIS TIME FORWARD, THIS CACHE LINE WILL SERVICE .. US
        if ( debugging > 1 ) fprintf ( stderr, "slot %u is being populated\n", ( hval % REGEX_CACHE_SLOTS ) );
        if ( rec->pattern != NULL ) free      ( rec->pattern );
        if ( rec->func    != NULL ) pcre_free ( rec->func );
        if ( rec->postcompile != NULL ) 
          {
            pcre_free ( rec->postcompile );
            rec->postcompile = NULL;
	  }
        rec->pattern = strdup ( pattern );
        rec->hashvalue = hval;
        rec->func = pcre_compile ( 
              pattern, 
              PCRE_DOLLAR_ENDONLY | PCRE_DOTALL,
              &err,
              &erroffset,
              NULL );
        if ( debugging > 1 ) fprintf ( stderr, "func = 0x%X\n", rec->func );
        if ( rec->func == NULL ) 
          {
            free ( rec->pattern ); 
            rec->pattern = NULL;
            FEXIT  return ( -6 ); // FIXME -- report back cause of error
	  }
        if ( regex_postcompile > 0 ) 
          {
            if ( debugging > 1 ) fprintf ( stderr, "running postcompile\n" );
            err = NULL;
            rec->postcompile = pcre_study ( rec->func, 0, &err );
            if ( debugging > 1 ) if ( err != NULL ) fprintf ( stderr, "postcompile error: (0x%X) '%s'\n", err, err );
	  } else if ( debugging > 1 ) fprintf ( stderr, "skipping postcompile\n" );
        if ( debugging > 1 ) fprintf ( stderr, "pcomp = 0x%X\n", rec->postcompile );
      } // end of putting stuff in cache line
    if ( debugging > 1 ) fprintf ( stderr, "running pcre_exec with 0x%X 0x%X\n", rec->func, rec->postcompile );
    gotmatch = pcre_exec ( rec->func, rec->postcompile, s, strlen(s), 0, 0, vectors, REGEX_N_VECTORS );
    if ( debugging > 1 ) fprintf ( stderr, "result = %d\n", gotmatch );
    if ( ( gotmatch > 0 ) && ( x != NULL ) )
      {
        int i, res;
        const char *buf;
        res = 0;
        // first vector points at substring of complete match; ignoring
        for ( i = 1; ( i < REGEX_N_VECTORS ) && ( res == 0 ); i++ )
          {
            res = pcre_get_substring ( s, vectors, gotmatch, i, &buf );
            if ( res >= 0 ) 
              {
                res = 0; 
                xpt_add ( x, (char*) buf );
	      }
            else res = 1;
	  }
      } // end of adding matches to extant x
    if ( debugging > 1 ) fprintf ( stderr, "re() exiting\n" );
    if ( gotmatch < 0 ) gotmatch = 0;
    FEXIT  return ( gotmatch );
  } // end of function re

// zzapp -- define all these RECTL_* values in re.h
int re_ctl ( int parameter, int value )
  {
    switch ( parameter )
      {
        case 0 : return -1;
        case RECTL_GET_CACHE_REPLACEMENT_POLICY_DEFAULT : return REGEX_DEFAULT_REPLACEMENT_POLICY;
        case RECTL_GET_CACHE_REPLACEMENT_POLICY : return cache_replacement_policy;
        case RECTL_SET_CACHE_REPLACEMENT_POLICY : cache_replacement_policy = value; return 0;
        case RECTL_GET_POSTCOMPILE_DEFAULT : return REGEX_DEFAULT_POSTCOMPILE;
        case RECTL_GET_POSTCOMPILE : return regex_postcompile;
        case RECTL_SET_POSTCOMPILE : regex_postcompile = value; return 0;
        case RECTL_GET_CACHE_SLOTS : return REGEX_CACHE_SLOTS;
        case RECTL_GET_N_VECTORS : return REGEX_N_VECTORS;
        case RECTL_SET_DEBUG : debugging = value; return 0;
        case RECTL_GET_DEBUG : return REGEX_N_VECTORS;
      }
    return ( -2 );
  }

